
; ######################################
; #                                    #
; #       MESSAGE BOX EXPANSION        #
; #                                    #
; #  made in 2012-13 by WhiteYoshiEgg  #
; #                                    #
; ######################################


	; Refer to readme.html for instructions and further information.


header
lorom



; ###############
; # Definitions #
; ###############


	!YoshiMessage		= $00		; The message Yoshi displays when rescued ($00 - $7D)

	!Chnl			= 7		; The HDMA channel to use for the box window
						; (apparently needs to be the same as the one SMW uses,
						; which is 7 unless you have some DMA remapping patch applied)

	; The three RAM addresses below were only used by SMW's message boxes,
	; which this patch obsoletes.
	; You can change them if you want,
	; but it's perfectly fine to leave them as they are.


	!MessageBoxState	= $1426		; "Non-existent", "Opening", "Displaying text", "Closing"
						; (values #$00 - #$03 respectively)


	!MessageBoxTimer	= $1B88		; Timer for various message box related stuff


	!MessageToShow		= $1B89		; What message to display (#$00 - #$7D)









; ######################################################
; # Macro to help you specify which message goes where #
; ######################################################

macro setMsg(lvl,scr,msg)
	org MessageList+$<scr><lvl>
	db $<msg>
endmacro






; ###########
; # Hijacks #
; ###########

org $00A1DA				; \
	LDA !MessageBoxState		;  |
	BEQ +				;  | Interrupt the game
	autoclean JSL MessageBoxMain	;  | when there's a message box running
	RTS				;  |
	+				; /

org $038D87				; \  Hijack message box sprite to load
	autoclean JML GetMessage	; /  the number of the message to display

org $01EC3B				;    Display a custom Yoshi Message
	autoclean JML YoshiMessage

org $05D8B9				; \
	JSR LevelNum			;  | Saves the current level number to $010B,
org $05DC46				;  | just in case someone doesn't have uberASM
LevelNum:				;  | or something similar installed
	LDA $0E : STA $010B : ASL : RTS	; /




; ############
; # New Code #
; ############

freecode
prot MessageProperties, MessageList, Messages
reset bytes
print ""
print " Inserted at $",pc




; ############################
; # Main Message Box Routine #
; ############################

MessageBoxMain:

	PHB : PHK : PLB			; \
	REP #$20			;  | Prepare the data bank register
	LDA $00 : PHA			;  | and preserve the RAM we're working with
	LDA $02 : PHA			;  | just to be on the safe side
	LDA $04 : PHA			;  |
	SEP #$20			; /

	JSR .handleMessageBoxState	;    Let a subroutine do the actual work

	REP #$20			; \
	PLA : STA $04			;  |
	PLA : STA $02			;  | Get back the things we saved
	PLA : STA $00			;  |
	SEP #$20			;  |
	PLB				; /

	RTL				;    That's all, folks!




; ####################
; # Anzeigen der Box #
; ####################

.handleMessageBoxState

	LDA $71				; \
	CMP #$09			;  | Instantly remove the box window
	BNE .notDying			;  | when the player is dying
	JMP .closing_nextState		;  | (otherwise the death animation would stop
					;  | with the death music still playing)
.notDying				; /


	LDA !MessageBoxState		; \
	DEC				;  |
	JSL $0086DF			;  | Handle the message box state
	dw .opening			;  | (if we get here it's never zero anyway, so DEC)
	dw .there			;  |
	dw .closing			; /




	; #$01: Opening

.opening

	LDA #$22			; \
	STA $41				;  |
	STA $43				;  | Change the screen settings
	STA $44				;  | and enable HDMA
	STZ $42				;  |
	LDA #$1<<!Chnl			;  |
	TSB $0D9F			; /

	LDA !MessageBoxTimer		; \
	CMP #23				;  | Change the state after 22 frames
	BEQ ..nextState			; /

	; hooray, math~

	TAX				; \
	LDA.w .boxWidths,x		;  | Save current window with and height
	STA $00				;  | in $00-$01
	LDA.w .boxHeights,x		;  |
	STA $01				; /

	LDA #$FF			; \
	SEC : SBC $00 : LSR		;  | Save horizontal start and end points
	STA $02				;  | in $02-$03
	CLC : ADC $00			;  |
	STA $03				; /

	LDA #$B0			; \  Determine the vertical starting point
	SEC : SBC $01			;  | (=index to the windowing HDMA table at $04A0)
	AND #$FE			;  | (making sure it's an even number too)
	TAX				; /  #$B0 is basically the box center's Y position times two

	REP #$10			; \
..loop					;  |
	LDA $02				;  | Write the horizontal start and end points to $04A0+
	STA $04A0,x			;  | using the box height as the loop counter
	LDA $03				;  |
	STA $04A1,x			;  |
	INX #2				;  |
	DEC $01				;  |
	LDA $01				;  |
	BNE ..loop			;  |
	SEP #$10			; /

	INC !MessageBoxTimer		;    Increment the timer

	RTS				;    End of state #$01


..nextState				; \  When we're finished,
	STZ !MessageBoxTimer		;  | reset the timer
	INC !MessageBoxState		; /  and increment the box state

	RTS				;    End of state #$01






	; #$02: Displaying text

.there

	LDA !MessageBoxTimer		; \  If the timer has exceeded 9
	CMP #10				;  | (meaning all lines have been drawn),
	BCS ..waitForInput		; /  jump right to the "waiting for input" code

	LDA $13				; \  If the timer's still ticking
	AND #$01			;  | (meaning text is still being drawn),
	BEQ ..return			; /  do nothing every second frame

	JSR .displayText		; \  Draw the current line,
	INC !MessageBoxTimer		;  | increment the timer
	BRA ..return			; /  and skip the "waiting for input" code

..waitForInput
	LDA $15				; \
	AND #$F0			;  |
	BEQ ..return			;  | Increment the box state
	EOR $16				;  | when certain buttons are being pressed
	AND #$F0			;  |
	BEQ ..checkIntro		;  | (copied from SMW)
	LDA $17				;  |
	AND #$C0			;  |
	BEQ ..return			;  |
	EOR $18				;  |
	AND #$C0			;  |
	BNE ..return			; /

..checkIntro
	LDA $0109
	BNE ..doIntroStuff

..nextState
	JSR .eraseText			; \
	LDA #22				;  | When we're done, erase the text,
	STA !MessageBoxTimer		;  | set the timer to 22
	INC !MessageBoxState		;  | and increment the box state
	RTS				; /

..doIntroStuff
	LDA #$78			; \
	STA $1F19			;  | Intro stuff
	STZ $0109			;  | (I actually forgot what this is doing exactly,
	STZ $0DD5			;  | all I know is that it ensures the intro level won't mess up)
	LDA #$0B			;  |
	STA $0100			; /

..return
	RTS				;    End of state #$02






	; #$03: Closing

.closing

	LDA !MessageBoxTimer		; \  The timer counts down
	BEQ ..nextState			; /  from 22 to zero here

	REP #$30			; \
	LDX #448			;  |
-					;  | Remove existing windows
	LDA #$00FF			;  | before drawing the current one
	STA $04A0,x			;  | (necessary here since every window is
	DEX #2				;  | smaller than the preceding one)
	BPL -				;  |
	SEP #$30			; /

	LDA !MessageBoxTimer		; \
	TAX				;  |
	LDA.w .boxWidths,x		;  | Same math the "opening" code uses,
	STA $00				;  | only with the box dimensions
	LDA.w .boxHeights,x		;  | getting smaller each frame
	STA $01				;  | since the timer's reversed
					;  |
	LDA #$FF			;  |
	SEC : SBC $00 : LSR		;  |
	STA $02				;  |
	CLC : ADC $00			;  |
	STA $03				;  |
					;  |
	LDA #$B0			;  |
	SEC : SBC $01			;  |
	AND #$FE			;  |
	TAX				;  |
					;  |
	REP #$10			;  |
..loop					;  |
	LDA $02				;  |
	STA $04A0,x			;  |
	LDA $03				;  |
	STA $04A1,x			;  |
	INX #2				;  |
	DEC $01				;  |
	LDA $01				;  |
	BNE ..loop			;  |
	SEP #$10			; /

	DEC !MessageBoxTimer		;    Decrement timer

	RTS				;    End of state #$03


..nextState
	STZ !MessageBoxTimer		; \  Reset box timer and state
	STZ !MessageBoxState		; /

	REP #$30			; \
	LDX #448			;  |
-					;  |
	LDA #$00FF			;  | Reset HDMA table
	STA $04A0,x			;  | (= remove all windowing)
	DEX #2				;  |
	BPL -				;  |
	SEP #$30			; /

	LDA #$1<<!Chnl			; \  Disable HDMA
	TRB $0D9F			; /

	RTS






; #########################################
; # Box widths and heights for each frame #
; # during opening/closing                #
; #########################################

.boxWidths
	db 4,14,24,34,44,54,64,74,84,94,104,114,124,134,144,154,164,174,184,194,204,214,224

.boxHeights
	db 8,12,16,20,24,28,32,36,40,44,48,52,56,60,64,68,72,76,80,84,88,92,96






; ######################
; # Message displaying #
; ######################

.displayText

	; Uses:
	; $00-$02:	Address of the tilemap for the current line being shown
	; $03:		This message's YXPCCCTT properties
	; $04:		Current line's Y position
	; X:		Index to $7F837D
	; Y:		Loop counter (current character in the line)



	REP #$20

	LDA !MessageToShow		; \
	ASL #2				;  |
	STA $00				;  | Multiply the number of the message to show with 260
	ASL #6				;  | and add that to the beginning of the message table
	CLC : ADC $00			;  | (this determines the address of the current message)
					;  |
	CLC : ADC.w #Messages		;  |
	STA $00				; /

	LDA !MessageBoxTimer		; \
	AND #$00FF			;  |
	ASL				;  | Multiply the timer by the number of characters per line
	STA $03				;  | and add the address of the current message
	ASL #2				;  | (this determines the address of the current line to show)
	STA $05				;  |
	ASL				;  |
	CLC : ADC $03			;  |
	CLC : ADC $05			; /

	CLC : ADC $00			; \  Save the above to $00-$01
	STA $00				; /

	SEP #$20			; \  Save the bank to $02
	LDA.b #Messages>>16		;  | (all the above makes [$00] the address of
	STA $02				; /  the tilemap for the current line to show)

	LDX !MessageToShow		; \
	LDA.l MessageProperties,x	;  | Save tile properties to 03
	STA $03				; /

	REP #$30			; \
	LDA $7F837B : TAX		;  | X = index to $7F837D,
	LDY #$0000			;  | Y = 0
	SEP #$20			; /

	LDA !MessageBoxTimer		; \
	CLC : ADC #$20			;  | Save current line's Y position to $04
	STA $04				; /


	; This saves the current line's stripe image data
	; to the layer 3 tilemap table at $7F837D

..header

	LDA $04				; \
	AND #$20			;  |
	LSR #2				;  | Stripe image header, Byte 1
	STA $7F837D,x			;  |
	LDA $04				;  | (bitwise shenanigans galore
	AND #$18			;  | to cram the Y position in there)
	LSR #3				;  |
	ORA $7F837D,x			;  |
	ORA #$50			;  |
	STA $7F837D,x			;  |
	INX				; /

	LDA $04				; \
	ASL #5				;  |
	AND #$E0			;  | Byte 2
	ORA #$03			;  | (X position, remaining bits of Y position)
	STA $7F837D,x			;  |
	INX				; /

	LDA #$00			; \
	STA $7F837D,x			;  | Byte 3 (Direction, RLE, some length bits)
	INX				; /

	LDA #51				; \
	STA $7F837D,x			;  | Byte 4 (Length = [26 characters] * [2 bytes per character] - 1 = 51)
	INX				; /


..body

...loop
	LDA [$00],y			; \
	STA $7F837D,x			;  | Save the current character
	INX				; /
	LDA $03				; \
	STA $7F837D,x			;  | Save the property byte right after
	INX				; /

	INY				; \
	CPY.w #26			;  | ^ do so 26 times, for each character in the line
	BNE ...loop			; /

..return				;    At the end of the line,

	LDA #$FF			; \  put an FF byte
	STA $7F837D,x			; /
	REP #$20			; \  and save the new index
	TXA : STA $7F837B		; /
	SEP #$30

	STZ $12				;    Tell SMW to upload an image from $7F837D
	RTS				;    done~






; ####################
; # Message removing #
; ####################

.eraseText

	REP #$30			; \
	LDA $7F837B : TAX		;  |
	LDY.w #$07			;  | Put a stripe image with empty tiles
	SEP #$20			;  | to overwrite the message
..loop					;  | at $7F837D
	LDA ..emptyTilemap,x		;  |
	STA $7F837D,x			;  |
	INX				;  |
	DEY				;  |
	BPL ..loop			;  |
					;  |
	REP #$20			;  |
	TXA : STA $7F837B		;  |
	SEP #$30			;  |
					;  |
	STZ $12				;  |
	RTS				; /

..emptyTilemap
	db $58,$03,$42,$72
	db $FC,$38
	db $FF





; #########################################
; # Get the number of the message to show #
; #########################################

GetMessage:

	PHX

	LDA $14E0,x			; \
	XBA				;  |
	LDA $010B			;  | Save the screen number and the low byte of the level number
	REP #$30			;  | to 16-bit X
	TAX				;  |
	SEP #$20			; /

	LDA MessageList,x		; \  Save the number of the corresponding message
	STA !MessageToShow		; /

	SEP #$10
	PLX

	LDA #$01			; \  Tell the message box to show up
	STA !MessageBoxState		; /

	JML $038D93			;    Return










; ###################
; # Yoshi's message #
; ###################

YoshiMessage:

	LDA #!YoshiMessage
	STA !MessageToShow

	LDA #$01
	STA !MessageBoxState

	JML $01EC40



print " ",bytes," bytes used."













; ################
; # Various data #
; ################

freedata
reset bytes
print " Data inserted at $",pc






; #########################################
; # YXPCCCTT properties for every message #
; #########################################

MessageProperties:

	db $39,$39,$39,$39,$39,$39,$39,$39	; $00-$07
	db $39,$39,$39,$39,$39,$39,$39,$39	; $08-$0F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $10-$17
	db $39,$39,$39,$39,$39,$39,$39,$39	; $18-$1F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $20-$27
	db $39,$39,$39,$39,$39,$39,$39,$39	; $28-$2F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $30-$37
	db $39,$39,$39,$39,$39,$39,$39,$39	; $38-$3F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $40-$47
	db $39,$39,$39,$39,$39,$39,$39,$39	; $48-$4F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $50-$57
	db $39,$39,$39,$39,$39,$39,$39,$39	; $58-$5F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $60-$67
	db $39,$39,$39,$39,$39,$39,$39,$39	; $68-$6F
	db $39,$39,$39,$39,$39,$39,$39,$39	; $70-$77
	db $39,$39,$39,$39,$39,$39		; $78-$7D






; #####################################################
; # The table determining which message to show where #
; #####################################################

MessageList:

	rep 8192 : db $00

	print " ",bytes," bytes used."

	incsrc message_list.asm









; ########################
; # Actual message texts #
; ########################

freedata
reset bytes
print " Messages inserted at $",pc


Messages:

	cleartable : table ascii.txt
	incsrc messages.asm



print " ",bytes," bytes used."
print ""